package uniq

  1. Overview
  2. Docs

Module Uniq_mlSource

Sourcetype Effect.t +=
  1. | Read_file : Read.kind * string * Namespaced.t -> Comp_unit.s Effect.t
Sourceexception Invalid_source_file of string
Sourceval run : ?version:Sys.ocaml_release_info -> ?stdlib:bool -> Name.t list -> Comp_unit.u list Comp_unit.pair
Sourceval run_into : ?version:Sys.ocaml_release_info -> ?stdlib:bool -> current:Fpath.t -> Name.t list -> Comp_unit.u list Comp_unit.pair

run_into ?version ~current filenames it resolves the dependencies of given source files filenames in a given folder current (considered to be the root for the namespace).

Unlike run, this function manages the constr:Read_file effect.