package uniq

  1. Overview
  2. Docs

Module Uniq_modSource

Sourcetype k = [
  1. | `All
  2. | `Intf
  3. | `Impl
]
Sourcetype t = [
  1. | `All
  2. | `Sources
  3. | `Objects
]
Sourcetype c = [
  1. | `All
  2. | `Bytecode
  3. | `Native
]

search ?filters ~roots p digest searches for an OCaml object according to a module path and an optional signature. The result can be filtered by the type of object using the filters option. We only searches in the folders specified by roots.